@@ -40,15 +40,17 @@
 #include "src/scan.h"
 #include "src/dbus.h"
 #include "src/agent.h"
-#include "src/device.h"
+#include "src/netdev.h"
 #include "src/wiphy.h"
+#include "src/watchlist.h"
+#include "src/station.h"
 #include "src/eap.h"
 #include "src/knownnetworks.h"
 #include "src/network.h"
 
 struct network {
 	char *object_path;
-	struct device *device;
+	struct station *station;
 	struct network_info *info;
 	unsigned char *psk;
 	char *passphrase;
@@ -103,16 +105,16 @@ bool network_info_match(const void *a, const void *b)
 	return true;
 }
 
-static bool network_info_ptr_match(const void *a, const void *b)
-{
-	return a == b;
-}
-
 static bool network_secret_check_cacheable(void *data, void *user_data)
 {
 	struct eap_secret_info *secret = data;
 
-	return secret->cache_policy == EAP_CACHE_NEVER;
+	if (secret->cache_policy == EAP_CACHE_NEVER) {
+		eap_secret_info_free(secret);
+		return true;
+	}
+
+	return false;
 }
 
 void network_connected(struct network *network)
@@ -132,11 +134,12 @@ void network_connected(struct network *network)
 		/*
 		 * This is an open network seen for the first time:
 		 *
-		 * Write an empty settings file to keep track of the
+		 * Write a settings file to keep track of the
 		 * last connected time.  This will also make iwd autoconnect
 		 * to this network in the future.
 		 */
-		network->settings = l_settings_new();
+		if (!network->settings)
+			network->settings = l_settings_new();
 
 		storage_network_sync(network_get_security(network),
 					network->info->ssid,
@@ -273,20 +276,20 @@ static void network_info_put(struct network_info *network)
 	if (--network->seen_count)
 		return;
 
-	if (!networks)
+	if (!networks || network->is_known)
 		return;
 
-	if (l_queue_remove(networks, network))
-		network_info_free(network);
+	l_queue_remove(networks, network);
+	network_info_free(network);
 }
 
-struct network *network_create(struct device *device, const char *ssid,
+struct network *network_create(struct station *station, const char *ssid,
 				enum security security)
 {
 	struct network *network;
 
 	network = l_new(struct network, 1);
-	network->device = device;
+	network->station = station;
 	network->info = network_info_get(ssid, security);
 
 	network->bss_list = l_queue_new();
@@ -299,11 +302,6 @@ const char *network_get_ssid(const struct network *network)
 	return network->info->ssid;
 }
 
-struct device *network_get_device(const struct network *network)
-{
-	return network->device;
-}
-
 const char *network_get_path(const struct network *network)
 {
 	return network->object_path;
@@ -421,6 +419,10 @@ static int network_load_psk(struct network *network)
 		return -EINVAL;
 	}
 
+	l_free(network->passphrase);
+	network->passphrase = l_settings_get_string(network->settings,
+						"Security", "Passphrase");
+
 	return 0;
 }
 
@@ -438,10 +440,20 @@ void network_sync_psk(struct network *network)
 	l_settings_set_value(network->settings, "Security",
 						"PreSharedKey", hex);
 
+	if (network->passphrase)
+		l_settings_set_string(network->settings, "Security",
+							"Passphrase",
+							network->passphrase);
+
 	fs_settings = storage_network_open(SECURITY_PSK, network->info->ssid);
 	if (fs_settings) {
 		l_settings_set_value(fs_settings, "Security",
 						"PreSharedKey", hex);
+		if (network->passphrase)
+			l_settings_set_string(fs_settings, "Security",
+							"Passphrase",
+							network->passphrase);
+
 		storage_network_sync(SECURITY_PSK, network->info->ssid,
 					fs_settings);
 		l_settings_free(fs_settings);
@@ -454,7 +466,8 @@ void network_sync_psk(struct network *network)
 
 int network_autoconnect(struct network *network, struct scan_bss *bss)
 {
-	struct wiphy *wiphy = device_get_wiphy(network->device);
+	struct station *station = network->station;
+	struct wiphy *wiphy = station_get_wiphy(station);
 	bool is_autoconnectable;
 	bool is_rsn;
 	int ret;
@@ -475,6 +488,9 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 		return -ENOTSUP;
 	}
 
+	if (!network_settings_load(network))
+		return -ENOKEY;
+
 	if (is_rsn) {
 		struct ie_rsn_info rsn;
 
@@ -484,13 +500,18 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 		if (!wiphy_select_cipher(wiphy, rsn.pairwise_ciphers) ||
 				!wiphy_select_cipher(wiphy, rsn.group_cipher)) {
 			l_debug("Cipher mis-match");
-			return -ENETUNREACH;
+			ret = -ENETUNREACH;
+			goto close_settings;
+		}
+
+		if (rsn.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256 &&
+				!l_settings_has_key(network->settings,
+						"Security", "Passphrase")) {
+			ret = -ENOKEY;
+			goto close_settings;
 		}
 	}
 
-	if (!network_settings_load(network))
-		return -ENOKEY;
-
 	/* If no entry, default to Autoconnectable=True */
 	if (!l_settings_get_bool(network->settings, "Settings",
 					"Autoconnect", &is_autoconnectable))
@@ -522,7 +543,7 @@ int network_autoconnect(struct network *network, struct scan_bss *bss)
 			goto close_settings;
 	}
 
-	return __device_connect_network(network->device, network, bss);
+	return __station_connect_network(station, network, bss);
 
 close_settings:
 	network_settings_close(network);
@@ -581,7 +602,7 @@ struct scan_bss *network_bss_find_by_addr(struct network *network,
 struct scan_bss *network_bss_select(struct network *network)
 {
 	struct l_queue *bss_list = network->bss_list;
-	struct wiphy *wiphy = device_get_wiphy(network->device);
+	struct wiphy *wiphy = station_get_wiphy(network->station);
 	const struct l_queue_entry *bss_entry;
 
 	switch (network_get_security(network)) {
@@ -616,6 +637,7 @@ static void passphrase_callback(enum agent_result result,
 				void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
 	struct scan_bss *bss;
 
 	l_debug("result %d", result);
@@ -665,7 +687,7 @@ static void passphrase_callback(enum agent_result result,
 	 */
 	network->update_psk = true;
 
-	device_connect_network(network->device, network, bss, message);
+	station_connect_network(station, network, bss, message);
 	l_dbus_message_unref(message);
 	return;
 
@@ -673,11 +695,24 @@ err:
 	network_settings_close(network);
 }
 
+static bool bss_is_sae(struct scan_bss *bss)
+{
+	struct ie_rsn_info rsn;
+
+	memset(&rsn, 0, sizeof(rsn));
+	scan_bss_get_rsn_info(bss, &rsn);
+
+	if (rsn.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256)
+		return true;
+
+	return false;
+}
+
 static struct l_dbus_message *network_connect_psk(struct network *network,
 					struct scan_bss *bss,
 					struct l_dbus_message *message)
 {
-	struct device *device = network->device;
+	struct station *station = network->station;
 
 	l_debug("");
 
@@ -688,7 +723,14 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 
 	l_debug("ask_psk: %s", network->ask_psk ? "true" : "false");
 
-	if (network->ask_psk || !network->psk) {
+	/*
+	 * A legacy psk file may only contain the PreSharedKey entry. For SAE
+	 * networks the raw Passphrase is required. So in this case where
+	 * the psk is found but no passphrase, we ask the agent. In this case
+	 * the psk file will be re-written to contain the raw passphrase.
+	 */
+	if (network->ask_psk || !network->psk ||
+			(!network->passphrase && bss_is_sae(bss))) {
 		network->ask_psk = false;
 
 		network->agent_request =
@@ -699,7 +741,7 @@ static struct l_dbus_message *network_connect_psk(struct network *network,
 		if (!network->agent_request)
 			return dbus_error_no_agent(message);
 	} else
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 
 	return NULL;
 }
@@ -898,6 +940,7 @@ static struct l_dbus_message *network_connect_8021x(struct network *network,
 						struct scan_bss *bss,
 						struct l_dbus_message *message)
 {
+	struct station *station = network->station;
 	int r;
 	struct l_queue *missing_secrets = NULL;
 	struct l_dbus_message *reply;
@@ -930,7 +973,7 @@ static struct l_dbus_message *network_connect_8021x(struct network *network,
 			goto error;
 		}
 
-		device_connect_network(network->device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 
 		return NULL;
 	}
@@ -955,12 +998,12 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 						void *user_data)
 {
 	struct network *network = user_data;
-	struct device *device = network->device;
+	struct station *station = network->station;
 	struct scan_bss *bss;
 
 	l_debug("");
 
-	if (device_is_busy(device))
+	if (station_is_busy(station))
 		return dbus_error_busy(message);
 
 	/*
@@ -978,7 +1021,7 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 	case SECURITY_PSK:
 		return network_connect_psk(network, bss, message);
 	case SECURITY_NONE:
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 		return NULL;
 	case SECURITY_8021X:
 		if (!network_settings_load(network))
@@ -993,7 +1036,7 @@ static struct l_dbus_message *network_connect(struct l_dbus *dbus,
 void network_connect_new_hidden_network(struct network *network,
 						struct l_dbus_message *message)
 {
-	struct device *device = network->device;
+	struct station *station = network->station;
 	struct scan_bss *bss;
 	struct l_dbus_message *error;
 
@@ -1021,7 +1064,7 @@ void network_connect_new_hidden_network(struct network *network,
 		error = network_connect_psk(network, bss, message);
 		break;
 	case SECURITY_NONE:
-		device_connect_network(device, network, bss, message);
+		station_connect_network(station, network, bss, message);
 		return;
 	default:
 		error = dbus_error_not_supported(message);
@@ -1054,9 +1097,10 @@ static bool network_property_is_connected(struct l_dbus *dbus,
 					void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
 	bool connected;
 
-	connected = device_get_connected_network(network->device) == network;
+	connected = station_get_connected_network(station) == network;
 	l_dbus_message_builder_append_basic(builder, 'b', &connected);
 	return true;
 }
@@ -1067,9 +1111,11 @@ static bool network_property_get_device(struct l_dbus *dbus,
 					void *user_data)
 {
 	struct network *network = user_data;
+	struct station *station = network->station;
+	struct netdev *netdev = station_get_netdev(station);
 
 	l_dbus_message_builder_append_basic(builder, 'o',
-					device_get_path(network->device));
+						netdev_get_path(netdev));
 
 	return true;
 }
@@ -1089,6 +1135,22 @@ static bool network_property_get_type(struct l_dbus *dbus,
 	return true;
 }
 
+static bool network_property_get_known_network(struct l_dbus *dbus,
+					struct l_dbus_message *message,
+					struct l_dbus_message_builder *builder,
+					void *user_data)
+{
+	struct network *network = user_data;
+
+	if (!network->info->is_known)
+		return false;
+
+	l_dbus_message_builder_append_basic(builder, 'o',
+					known_network_get_path(network->info));
+
+	return true;
+}
+
 static void setup_network_interface(struct l_dbus_interface *interface)
 {
 	l_dbus_interface_method(interface, "Connect", 0,
@@ -1107,6 +1169,9 @@ static void setup_network_interface(struct l_dbus_interface *interface)
 
 	l_dbus_interface_property(interface, "Type", 0, "s",
 					network_property_get_type, NULL);
+
+	l_dbus_interface_property(interface, "KnownNetwork", 0, "o",
+				network_property_get_known_network, NULL);
 }
 
 bool network_register(struct network *network, const char *path)
@@ -1181,18 +1246,14 @@ int network_rank_compare(const void *a, const void *b, void *user)
 	return network->rank - new_network->rank;
 }
 
-void network_rank_update(struct network *network)
+void network_rank_update(struct network *network, bool connected)
 {
-	bool connected;
-	int rank;
-
 	/*
 	 * Theoretically there may be difference between the BSS selection
 	 * here and in network_bss_select but those should be rare cases.
 	 */
 	struct scan_bss *best_bss = l_queue_peek_head(network->bss_list);
-
-	connected = device_get_connected_network(network->device) == network;
+	int rank;
 
 	/*
 	 * The rank should separate networks into four groups that use
@@ -1215,9 +1276,7 @@ void network_rank_update(struct network *network)
 			n = L_ARRAY_SIZE(rankmod_table) - 1;
 
 		rank = rankmod_table[n] * best_bss->rank + USHRT_MAX;
-	} else if (!l_queue_find(networks, network_info_ptr_match,
-					network->info))
-		/* Is a known network */
+	} else if (network->info->is_known)
 		rank = best_bss->rank;
 	else
 		rank = (int) best_bss->rank - USHRT_MAX; /* Negative rank */
@@ -1225,6 +1284,21 @@ void network_rank_update(struct network *network)
 	network->rank = rank;
 }
 
+static void emit_known_network_changed(struct station *station, void *user_data)
+{
+	struct network_info *info = user_data;
+	struct network *network;
+
+	network = station_network_find(station, info->ssid, info->type);
+	if (!network)
+		return;
+
+	l_dbus_property_changed(dbus_get_bus(),
+				network_get_path(network),
+				IWD_NETWORK_INTERFACE,
+				"KnownNetwork");
+}
+
 struct network_info *network_info_add_known(const char *ssid,
 						enum security security)
 {
@@ -1235,30 +1309,38 @@ struct network_info *network_info_add_known(const char *ssid,
 	search.type = security;
 
 	network = l_queue_remove_if(networks, network_info_match, &search);
-	if (network)
+	if (network) {
+		/* Promote network to is_known */
+		network->is_known = true;
+		station_foreach(emit_known_network_changed, network);
 		return network;
+	}
 
 	network = l_new(struct network_info, 1);
 	strcpy(network->ssid, ssid);
 	network->type = security;
+	network->is_known = true;
 
 	return network;
 }
 
-static void network_info_check_device(struct device *device, void *user_data)
+static void disconnect_no_longer_known(struct station *station, void *user_data)
 {
 	struct network_info *info = user_data;
 	struct network *network;
 
-	network = device_get_connected_network(device);
+	network = station_get_connected_network(station);
 
 	if (network && network->info == info)
-		device_disconnect(device);
+		station_disconnect(station);
 }
 
 void network_info_forget_known(struct network_info *network)
 {
-	__iwd_device_foreach(network_info_check_device, network);
+	network->is_known = false;
+
+	station_foreach(emit_known_network_changed, network);
+	station_foreach(disconnect_no_longer_known, network);
 
 	/*
 	 * Network is no longer a Known Network, see if we still need to
