@@ -2,7 +2,7 @@
  *
  *  Wireless daemon for Linux
  *
- *  Copyright (C) 2013-2014  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2013-2018  Intel Corporation. All rights reserved.
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -44,6 +44,7 @@
 #include "src/wiphy.h"
 #include "src/storage.h"
 #include "src/util.h"
+#include "src/common.h"
 
 static struct l_genl_family *nl80211 = NULL;
 static struct l_hwdb *hwdb;
@@ -91,6 +92,59 @@ enum ie_rsn_cipher_suite wiphy_select_cipher(struct wiphy *wiphy, uint16_t mask)
 	return 0;
 }
 
+enum ie_rsn_akm_suite wiphy_select_akm(struct wiphy *wiphy,
+					struct scan_bss *bss)
+{
+	struct ie_rsn_info info;
+	enum security security;
+
+	memset(&info, 0, sizeof(info));
+	scan_bss_get_rsn_info(bss, &info);
+
+	security = security_determine(bss->capability, &info);
+
+	/*
+	 * If FT is available, use FT authentication to keep the door open
+	 * for fast transitions.  Otherwise use SHA256 version if present.
+	 */
+	if (security == SECURITY_8021X) {
+		if ((info.akm_suites & IE_RSN_AKM_SUITE_FT_OVER_8021X) &&
+				bss->rsne && bss->mde_present)
+			return IE_RSN_AKM_SUITE_FT_OVER_8021X;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_8021X_SHA256)
+			return IE_RSN_AKM_SUITE_8021X_SHA256;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_8021X)
+			return IE_RSN_AKM_SUITE_8021X;
+	} else if (security == SECURITY_PSK) {
+		/*
+		 * Prefer connecting to SAE/WPA3 network, but only if SAE is
+		 * supported. This allows us to connect to a hybrid WPA2/WPA3
+		 * AP even if SAE/WPA3 is not supported.
+		 */
+		if (info.akm_suites & IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256 &&
+				wiphy_has_feature(wiphy, NL80211_FEATURE_SAE))
+			return IE_RSN_AKM_SUITE_FT_OVER_SAE_SHA256;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_SAE_SHA256 &&
+				wiphy_has_feature(wiphy, NL80211_FEATURE_SAE))
+			return IE_RSN_AKM_SUITE_SAE_SHA256;
+
+		if ((info.akm_suites & IE_RSN_AKM_SUITE_FT_USING_PSK) &&
+				bss->rsne && bss->mde_present)
+			return IE_RSN_AKM_SUITE_FT_USING_PSK;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_PSK_SHA256)
+			return IE_RSN_AKM_SUITE_PSK_SHA256;
+
+		if (info.akm_suites & IE_RSN_AKM_SUITE_PSK)
+			return IE_RSN_AKM_SUITE_PSK;
+	}
+
+	return 0;
+}
+
 static void wiphy_free(void *data)
 {
 	struct wiphy *wiphy = data;
@@ -182,6 +236,17 @@ bool wiphy_can_connect(struct wiphy *wiphy, struct scan_bss *bss)
 		if (rsn_info.mfpr && !wiphy_select_cipher(wiphy,
 					rsn_info.group_management_cipher))
 			return false;
+
+		/*
+		 * if the AP ONLY supports SAE/WPA3, then we can only connect
+		 * if the wiphy feature is supported. Otherwise the AP may list
+		 * SAE as one of the AKM's but also support PSK (hybrid). In
+		 * this case we still want to allow a connection even if SAE
+		 * is not supported.
+		 */
+		if (IE_AKM_IS_SAE(rsn_info.akm_suites) &&
+				!wiphy_has_feature(wiphy, NL80211_FEATURE_SAE))
+			return false;
 	} else if (r != -ENOENT)
 		return false;
 
